
#include <ssvd.h>
#include <math.h>
#include <config.h>
#include <assert.h>
#include <sys/param.h>

#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_rng.h>

#include <lapacke.h>

/*
 * Create a random complex matrix with the real and imaginary parts uniformly
 * distributed between low and high.
 * */
int ssvd_random_matrix(gsl_matrix_complex * A, double low, double high)
{
    size_t m = A->size1;
    size_t n = A->size2;

	/* Init random number generator */
	gsl_rng_env_setup();
	const gsl_rng_type *T = gsl_rng_default;
	gsl_rng *r = gsl_rng_alloc(T);
	gsl_complex rnd_tmp;

	/* reuse high as scaling variable for random dist */
	high = high - low;

	/* Fill matrix with random complex numbers */
	for (size_t i=0; i<m; i++) {
        for (size_t j=0; j<n; j++) {
		    rnd_tmp = GSL_COMPLEX_ZERO;
		    rnd_tmp = gsl_complex_rect(gsl_rng_uniform(r) * high + low,
		    			   gsl_rng_uniform(r) * high + low);

		    gsl_matrix_complex_set(A, i, j, rnd_tmp);
        }
	}
}

/* Creates a double complex test (m,n) test matrix with exponentially
 * decaying singular values 
 *
 * A:
 *      The output matrix to hold the test data
 *
 * scale:
 *      Value of the largest singular value
 *
 * decay_const:
 *      The i-th singular value is scale * exp(-i * decay_const)
 *
 * */
int ssvd_eigen_matrix(gsl_matrix_complex *A, double scale, double decay_const)
{

    size_t m = A->size1;
    size_t n = A->size2;
    size_t lda = A->tda;
    size_t mn = MIN(m, n);
    size_t kl = m-1;
    size_t ku = n-1;
    int iseed[4] = {2, 300, 95, 666};

    /* Calculate singular values */
    double *diag = calloc(MIN(m,n), sizeof(double));
    for (size_t i=0; i<MIN(m,n); i++) {
        diag[i] = scale * exp(-decay_const * (double) i);
    }

    lapack_complex_double *a = (lapack_complex_double *)
        gsl_matrix_complex_ptr(A, 0, 0);
    LAPACKE_zlagge(LAPACK_ROW_MAJOR, m, n, kl, ku, diag, a, lda, iseed);

}

/*
 * A wrapper of the LAPACKE_zgeqrf complex QR decomposition working on complex
 * gsl matrices. Note that in contrast to the LAPACK routine, this routine
 * calculates the (m, min(m,n)) unitary matrix Q explicitly on request instead of
 * returning the corresponding Householder transforms. As a result the matrix
 * R is lost if calculate_Q=1.
 * 
 * :param A:    On input the complex (m,n) matrix A. On output the results of
 *              the LAPACK zgeqrf routine if calculate_Q=0. Else the first
 *              min(m,n) columns of A contain Q.
 *
 * :param calculate_Q:  Explicitly calculates Q instead of returning the
 *                      Householder transforms.
 */
int ssvd_zgeqrf(gsl_matrix_complex * A, int calculate_Q)
{
	size_t m = A->size1;
	size_t n = A->size2;
	size_t lda = A->tda;
    size_t mn = MIN(m, n);

	/* Allocate space for Housholder vector tau */
	lapack_complex_double *tau = calloc(mn, sizeof(lapack_complex_double));

	/* Get correct pointer type from gsl matrix struct */
	lapack_complex_double *a = (lapack_complex_double *)
	    gsl_matrix_complex_ptr(A, 0, 0);

	/* QR decomposition - results as series of Householder matrices */
	LAPACKE_zgeqrf(LAPACK_ROW_MAJOR, m, n, a, lda, tau);

	/* Recover Q from result of QR decomposition */
    if (calculate_Q == 1) {
	    LAPACKE_zungqr(LAPACK_ROW_MAJOR, m, mn, mn, a, lda, tau);
    }
	free(tau);
}

/*
 * A wrapper of the LAPACKE zgesvd complex singular value decomposition working
 * on complex gsl matrices. This routine calculates the compact SVD A = U S Va.
 *
 * :param A:    A complex (m,n) gsl matrix to be decomposed.
 *
 * :param U:    A complex (m, min(m,n)) gsl matrix. Contains the left
 *              eigenvectors as columns on output.
 *
 * :param S:    A double real gsl vector of length min(m,n), containing the
 *              singular values on output.
 *
 * :param Va:   A complex (min(m,n) n) gsl matrix. Contains the right
 *              eigenvectors as rows on output.
 *
 * */
int ssvd_zgesvd(gsl_matrix_complex * A, gsl_matrix_complex * U, gsl_matrix_complex
	   * Va, gsl_vector * S)
{
	size_t m = A->size1;
	size_t n = A->size2;
	size_t lda = A->tda;
	size_t ldu = MIN(m,n);
	size_t ldva = n;

	/* Check input */
	assert(S->size == MIN(m, n));
	assert(U->size1 == m);
	assert(U->size2 == MIN(m, n));
	assert(Va->size1 == MIN(m, n));
	assert(Va->size2 == n);

	/* Allocate storage and pointers to data */
	lapack_complex_double *a = (lapack_complex_double *)
	    gsl_matrix_complex_ptr(A, 0, 0);
	lapack_complex_double *u = (lapack_complex_double *)
	    gsl_matrix_complex_ptr(U, 0, 0);
	lapack_complex_double *va = (lapack_complex_double *)
	    gsl_matrix_complex_ptr(Va, 0, 0);
	double *s = (double *)gsl_vector_ptr(S, 0);
	double *superb = calloc(MIN(m, n) - 1, sizeof(double));

	LAPACKE_zgesvd(LAPACK_ROW_MAJOR, 'S', 'S', m, n, a, lda, s, u, ldu, va,
		       ldva, superb);

	free(superb);
}

/*
 * Calculate the normalization constant sum(|A_ij|^2) for a complex matrix A.
 * */
int ssvd_matrix_normalization(gsl_matrix_complex * A, double *norm)
{
	const size_t m = A->size1;
	const size_t n = A->size2;
	const size_t N = m * n;

	gsl_complex *a = gsl_matrix_complex_ptr(A, 0, 0);
	*norm = 0.0;
	for (size_t i = 0; i < N; i++) {
		*norm += gsl_complex_abs2(a[i]);
	}
}

/*
 * Compute the rank-k SVD of A. This routine is a more or less direct port of
 * the fbpca.py code by XX. This algorithm does not support centering of the
 * data (useful for PCA, not required for a pure SVD). The original
 * documentation is excellent and therefore reproduced below:
 *
 * Constructs a nearly optimal rank-k approximation U diag(s) Va to A,
 * centering the columns of A first when raw is False, using n_iter
 * normalized power iterations, with block size l, started with a
 * min(m,n) x l random matrix, when A is m x n; the reference PCA_
 * below explains "nearly optimal." k must be a positive integer <=
 * the smaller dimension of A, n_iter must be a nonnegative integer,
 * and l must be a positive integer >= k.

 * The rank-k approximation U diag(s) Va comes in the form of a
 * singular value decomposition (SVD) -- the columns of U are
 * orthonormal, as are the rows of Va, and the entries of s are all
 * nonnegative and nonincreasing. U is m x k, Va is k x n, and
 * len(s)=k, when A is m x n.

 * Increasing n_iter or l improves the accuracy of the approximation
 * U diag(s) Va; the reference PCA_ below describes how the accuracy
 * depends on n_iter and l. Please note that even n_iter=1 guarantees
 * superb accuracy, whether or not there is any gap in the singular
 * values of the matrix A being approximated, at least when measuring
 * accuracy as the spectral norm || A - U diag(s) Va || of the matrix
 * A - U diag(s) Va (relative to the spectral norm ||A|| of A, and
 * accounting for centering when raw is False).
 *
 * Note that this routine does not return the rank-k approximation directly
 * but the full matrices that include l-k more elements than the rank-k
 * approximation. The corresponding rank-k matrices are easily constructed by
 * creating appropriate submatrix views. A good starting value for l would be
 * l = k + 2
 *
 * :param A:    A complex (m,n) matrix
 * :param U:    A complex (m,l) matrix
 * :param Va:   A complex (l,n) matrix
 * :param S:    A real vector of length l
 * :param k:    The rank of the approximation
 * :param l:    Oversampling parameter l>k
 *
 * */
int ssvd_ssvd_work(gsl_matrix_complex * A, size_t k, size_t l, int n_iter,
        gsl_matrix_complex *U, gsl_matrix_complex *Va, gsl_vector *S)
{
	/* Bookkeeping information */
	size_t m = A->size1;
	size_t n = A->size2;
	double norm = 0;

	gsl_complex alpha = GSL_COMPLEX_ONE;
	gsl_complex beta = GSL_COMPLEX_ZERO;

	/* Check input */
    assert(U->size1 == m);
    assert(U->size2 == l);
    assert(Va->size1 == l);
    assert(Va->size2 == n);
    assert(S->size == l);
	assert(k > 0);
	assert(k <= MIN(m, n));
	assert(n_iter >= 0);
	assert(l >= k);

	/* Calculate matrix normalization for scaled eigenvalues 0...1 */
	// ssvd_matrix_normalization(A, &norm);

    /* Intermediate storage */
	gsl_matrix_complex *U_tmp = gsl_matrix_complex_calloc(l, l);
    gsl_matrix_complex *Q = gsl_matrix_complex_calloc(m, l);
    gsl_matrix_complex *R = gsl_matrix_complex_calloc(n, l);

    /* Try to reuse some memory */
    gsl_matrix_complex_view B_view = gsl_matrix_complex_view_array(R->data, l, n);
	gsl_matrix_complex *B = &(B_view.matrix);

    if (m >= n) {
	    /* Calculate Q by multiplication of A with random matrix */
	    ssvd_random_matrix(R, -1.0, 1.0);
	    gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, A, R, beta, Q);

	    /* Orthogonal Matrix Q with columns a basis for the old Q */
	    ssvd_zgeqrf(Q, 1);

	    /* Normalized power iterations */
	    for (int i = 0; i < n_iter; i++) {
	    	gsl_blas_zgemm(CblasConjTrans, CblasNoTrans, alpha, A, Q, beta,
	    		       R);
	        ssvd_zgeqrf(R, 1);

	    	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, A, R, beta,
	    		       Q);
	        ssvd_zgeqrf(Q, 1);
	    }

	    /* Construct B=Q*A with A \approx QB */
	    gsl_blas_zgemm(CblasConjTrans, CblasNoTrans, alpha, Q, A, beta, B);
	    ssvd_zgesvd(B, U_tmp, Va, S);
	    gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, Q, U_tmp, beta, U);
	     
    }
    else {
        /* Not implemented and unlikely to happen soon but in principle
         * trivial */
    }


    /* Free temporary storage */
    gsl_matrix_complex_free(U_tmp);
    gsl_matrix_complex_free(Q);
    gsl_matrix_complex_free(R);
}


/*
 * A convenience wrapper around ssvd_ssvd_work, that allocates the necessary
 * storage and directly returns the rank-k approximations. Note that the
 * caller is responsible for deallocating U, Va and S.
 * */
int ssvd_ssvd(gsl_matrix_complex *A, size_t k, size_t l, size_t n_iter,
        gsl_matrix_complex **U, gsl_matrix_complex **Va, gsl_vector **S)
{
    size_t m = A->size1;
    size_t n = A->size2;

    /* Allocate storage including space for oversampling */
    *U = gsl_matrix_complex_alloc(m, l);
    *Va = gsl_matrix_complex_alloc(l, n);
    *S = gsl_vector_alloc(l);

    ssvd_ssvd_work(A, k, l, n_iter, *U, *Va, *S);

    /* Manipulate matrices to discard oversampling rows / cols */
    (**U).size2 = k;
    (**Va).size1 = k;
    (**S).size = k;

}

int ssvd_matrix_sum(gsl_matrix_complex * A, int axis)
{
}

