
#include <ssvd.h>
#include <lapacke.h>

#include <check.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>
#include <math.h>

#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>


/* Global test matrices */
size_t m; size_t n;
gsl_matrix_complex *test_matrix;
gsl_matrix_complex *U;
gsl_matrix_complex *Va;
gsl_vector *S;

/*----------------------------------------------------------------------------
 * Test fixtures
 * -------------------------------------------------------------------------*/

/* Populate test matrix */
void setup(void)
{
    test_matrix = gsl_matrix_complex_alloc(m, n);
    ssvd_eigen_matrix(test_matrix, 1.0, 1.0);
}

/* Tear down test matrix */
void teardown(void)
{
    gsl_matrix_complex_free(test_matrix);
}

/* Populate test matrix and allocate temporary storage for SVD */
void
setup_svd()
{
    setup();
    U = gsl_matrix_complex_alloc(m, MIN(m,n));
    Va = gsl_matrix_complex_alloc(MIN(m,n), n);
    S = gsl_vector_alloc(MIN(m,n));
}

/* Tear down SVD temporary storage and test matrix. */
void
teardown_svd()
{
    teardown();
    gsl_matrix_complex_free(U);
    gsl_matrix_complex_free(Va);
    gsl_vector_free(S);
}


/*----------------------------------------------------------------------------
 * Helper routines for testing.
 * -------------------------------------------------------------------------*/

/*
 * Check if the columns of a matrix are orthonormal.
 * */
int check_col_orthonormality(gsl_matrix_complex *A)
{
    size_t m = A->size1;
    size_t n = A->size2;
    gsl_complex dotc;

    /* Check orthonormality of columns */
    for (size_t i=0; i<n; i++) {
        for (size_t j=i; j<n; j++) {
            gsl_vector_complex_view C1 = gsl_matrix_complex_column(A, i);
            gsl_vector_complex_view C2 = gsl_matrix_complex_column(A, j);
            gsl_blas_zdotc(&(C1.vector), &(C2.vector), &dotc);
            if (i == j) {
                ck_assert(fabs(gsl_complex_abs(dotc) - 1.0) < 1e-14);
            }
            else {
                ck_assert(gsl_complex_abs(dotc) < 1e-14);
            }
        }
    }
}


/*
 * Check if the rows of a matrix are orthonormal.
 * */
int check_row_orthonormality(gsl_matrix_complex *A)
{
    size_t m = A->size1;
    size_t n = A->size2;
    gsl_complex dotc;
    gsl_vector_complex_view R1; 
    gsl_vector_complex_view R2; 

    /* Check orthonormality of columns */
    for (size_t i=0; i<m; i++) {
        for (size_t j=i; j<m; j++) {
            R1 = gsl_matrix_complex_row(A, i);
            R2 = gsl_matrix_complex_row(A, j);
            gsl_blas_zdotc(&(R1.vector), &(R2.vector), &dotc);
            if (i == j) {
                ck_assert(fabs(gsl_complex_abs(dotc) - 1.0) < 1e-14);
            }
            else {
                ck_assert(gsl_complex_abs(dotc) < 1e-14);
            }
        }
    }
}

/*----------------------------------------------------------------------------
 * Tests
 * -------------------------------------------------------------------------*/

/*
 * Test orthonormality of the columns of the complex QR decomposition
 * ssvd_zgeqrf for tall, square and wide matrices.
 * */
START_TEST(qr_orthonormality)
{
    size_t dim1 = 400;
    size_t dim2[3] = {300, 400, 500};
    gsl_matrix_complex_view Q;

    for (size_t i=0; i<3; i++) {
        m = dim1;
        n = dim2[i];
        setup();
        /* QR decomposition */
        ssvd_zgeqrf(test_matrix, 1);
        Q = gsl_matrix_complex_submatrix(test_matrix, 0, 0, m, MIN(m,n));
        check_col_orthonormality(&(Q.matrix));
        teardown();
    }
}
END_TEST

/*
 * Test for orthonormality of the columns of U and the rows of Va of the SVD
 * decomposition.
 * */
START_TEST(svd_orthonormality)
{
    size_t dim1 = 40;
    size_t dim2[3] = {30, 40, 50};

    for (size_t i=0; i<3; i++) {
        m = dim1;
        n = dim2[i];
        setup_svd();
        /* SVD decomposition */
        ssvd_zgesvd(test_matrix, U, Va, S);
        check_col_orthonormality(U);
        check_row_orthonormality(Va);
        teardown_svd();
    }
}
END_TEST

/* Check for expected eigenvalues of SVD */
START_TEST(svd_eigenvalues)
{
    size_t dim1 = 40;
    size_t dim2[3] = {30, 40, 50};

    for (size_t i=0; i<3; i++) {
        m = dim1;
        n = dim2[i];
        setup_svd();
        /* SVD decomposition */
        ssvd_zgesvd(test_matrix, U, Va, S);
        for (size_t j=0; j<MIN(m,n); j++) {
            ck_assert(fabs(gsl_vector_get(S, j) - exp(-(double) j)) < 1e-14);
        }
        teardown_svd();
    }

}
END_TEST

/*
 * Check if A = U S Va is true
 * */
START_TEST(svd_reconstruction)
{
}
END_TEST

/* Test if the SSVD eigenvalues are correct */
START_TEST(ssvd_eigenvalues)
{
    m = 40;
    n = 40;
    size_t k = 10;
    size_t l = 12;
    size_t n_iter = 2;
    setup();
    gsl_matrix_complex *U;
    gsl_matrix_complex *Va;
    gsl_vector *S;

    ssvd_ssvd(test_matrix, k, l, n_iter, &U, &Va, &S);
    for (size_t i=0; i<k; i++) {
        ck_assert(fabs(gsl_vector_get(S, i) - exp(-(double)i)) < 1e-12);

    }
    //printf("%zd\n", U->tda);
    ck_assert(1);
}
END_TEST

/*----------------------------------------------------------------------------
 * Check unit testing framework boilerplate
 * -------------------------------------------------------------------------*/

Suite *ssvd_suite(void)
{
    Suite *s = suite_create("SSVD suite");

    /* Define test cases for QR decomposition */
    TCase *tc_qr = tcase_create("QR");
    tcase_add_test(tc_qr, qr_orthonormality);
    suite_add_tcase(s, tc_qr);

    /* Define test cases for SVD decomposition */
    TCase *tc_svd = tcase_create("SVD");
    tcase_add_test(tc_svd, svd_orthonormality);
    tcase_add_test(tc_svd, svd_eigenvalues);
    suite_add_tcase(s, tc_svd);

    /* Define test cases for statistical SVD*/
    TCase *tc_ssvd = tcase_create("SSVD");
    tcase_add_test(tc_ssvd, ssvd_eigenvalues);
    suite_add_tcase(s, tc_ssvd);


    return(s);

}


int
main(void)
{
    int number_failed;
    Suite *s = ssvd_suite();
    SRunner *sr = srunner_create(s);
    srunner_run_all(sr, CK_VERBOSE);
    number_failed = srunner_ntests_failed(sr);
    srunner_free(sr);
    return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}



